@@ -35,6 +35,12 @@ static GHashTable *provider_hash = NULL;
 
 static GSList *driver_list = NULL;
 
+struct connman_route {
+	char *host;
+	char *netmask;
+	char *gateway;
+};
+
 struct connman_provider {
 	struct connman_element element;
 	struct connman_service *vpn_service;
@@ -44,6 +50,7 @@ struct connman_provider {
 	char *host;
 	char *dns;
 	char *domain;
+	GSList *route_list;
 	struct connman_provider_driver *driver;
 	void *driver_data;
 };
@@ -84,6 +91,9 @@ static int connman_provider_setup_vpn_ipv4(struct connman_provider *provider,
 	g_free(element->ipv4.address);
 	element->ipv4.address = g_strdup(provider->element.ipv4.address);
 
+	g_free(element->ipv4.peer);
+	element->ipv4.peer = g_strdup(provider->element.ipv4.peer);
+
 	g_free(element->ipv4.netmask);
 	element->ipv4.netmask = g_strdup(provider->element.ipv4.netmask);
 
@@ -185,12 +195,14 @@ int __connman_provider_connect(struct connman_provider *provider)
 	DBG("provider %p", provider);
 
 	g_free(provider->element.ipv4.address);
+	g_free(provider->element.ipv4.peer);
 	g_free(provider->element.ipv4.netmask);
 	g_free(provider->element.ipv4.gateway);
 	g_free(provider->element.ipv4.broadcast);
 	g_free(provider->element.ipv4.pac);
 
 	provider->element.ipv4.address = NULL;
+	provider->element.ipv4.peer = NULL;
 	provider->element.ipv4.netmask = NULL;
 	provider->element.ipv4.gateway = NULL;
 	provider->element.ipv4.broadcast = NULL;
@@ -256,6 +268,7 @@ static int set_connected(struct connman_provider *provider,
 		char *nameservers = NULL, *name = NULL;
 		const char *value;
 		char *second_ns;
+		GSList *list;
 		int err;
 
 		__connman_service_indicate_state(provider->vpn_service,
@@ -307,6 +320,15 @@ static int set_connected(struct connman_provider *provider,
 		g_free(nameservers);
 		g_free(name);
 
+		for (list = provider->route_list; list; list = list->next) {
+			struct connman_route *route = list->data;
+
+			connman_inet_add_network_route(provider->element.index,
+							route->host,
+							route->gateway,
+							route->netmask);
+			/* XXX Need to handle IPv6 too */
+		}
 	} else {
 		connman_element_unregister_children(&provider->element);
 		__connman_service_indicate_state(service,
@@ -360,6 +382,7 @@ static void unregister_provider(gpointer data)
 static void provider_destruct(struct connman_element *element)
 {
 	struct connman_provider *provider = element->private;
+	GSList *list;
 
 	DBG("provider %p", provider);
 
@@ -368,6 +391,15 @@ static void provider_destruct(struct connman_element *element)
 	g_free(provider->domain);
 	g_free(provider->identifier);
 	g_free(provider->dns);
+
+	for (list = provider->route_list; list; list = list->next) {
+		struct connman_route *route = list->data;
+
+		g_free(route->host);
+		g_free(route->netmask);
+		g_free(route->gateway);
+	}
+	g_slist_free(provider->route_list);
 }
 
 static void provider_initialize(struct connman_provider *provider)
@@ -390,6 +422,7 @@ static void provider_initialize(struct connman_provider *provider)
 	provider->dns = NULL;
 	provider->domain = NULL;
 	provider->identifier = NULL;
+	provider->route_list = NULL;
 }
 
 static struct connman_provider *connman_provider_new(void)
@@ -609,6 +642,9 @@ int connman_provider_set_string(struct connman_provider *provider,
 	} else if (g_str_equal(key, "Address") == TRUE) {
 		g_free(provider->element.ipv4.address);
 		provider->element.ipv4.address = g_strdup(value);
+	} else if (g_str_equal(key, "Peer") == TRUE) {
+		g_free(provider->element.ipv4.peer);
+		provider->element.ipv4.peer = g_strdup(value);
 	} else if (g_str_equal(key, "Netmask") == TRUE) {
 		g_free(provider->element.ipv4.netmask);
 		provider->element.ipv4.netmask = g_strdup(value);
@@ -661,12 +697,12 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 	if (service == NULL)
 		return;
 
-	ipconfig = __connman_service_get_ipconfig(service);
+	ipconfig = __connman_service_get_ip4config(service);
 
 	if (ipconfig == NULL) {
-		__connman_service_create_ipconfig(service, index);
+		__connman_service_create_ip4config(service, index);
 
-		ipconfig = __connman_service_get_ipconfig(service);
+		ipconfig = __connman_service_get_ip4config(service);
 		if (ipconfig == NULL) {
 			DBG("Couldnt create ipconfig");
 			goto done;
@@ -677,6 +713,21 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 	__connman_ipconfig_set_index(ipconfig, index);
 
 
+	ipconfig = __connman_service_get_ip6config(service);
+
+	if (ipconfig == NULL) {
+		__connman_service_create_ip6config(service, index);
+
+		ipconfig = __connman_service_get_ip6config(service);
+		if (ipconfig == NULL) {
+			DBG("Couldnt create ipconfig for IPv6");
+			goto done;
+		}
+	}
+
+	connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_OFF);
+	__connman_ipconfig_set_index(ipconfig, index);
+
 done:
 	provider->element.index = index;
 }
@@ -686,6 +737,30 @@ int connman_provider_get_index(struct connman_provider *provider)
 	return provider->element.index;
 }
 
+int connman_provider_append_route(struct connman_provider *provider,
+					const char *host, const char *netmask,
+					const char *gateway)
+{
+	struct connman_route *route;
+
+	route = g_try_new0(struct connman_route, 1);
+	if (route == NULL)
+		return -ENOMEM;
+
+	route->host = g_strdup(host);
+	route->netmask = g_strdup(netmask);
+	route->gateway = g_strdup(gateway);
+
+	provider->route_list = g_slist_append(provider->route_list, route);
+
+	return TRUE;
+}
+
+const char *connman_provider_get_driver_name(struct connman_provider *provider)
+{
+	return provider->driver->name;
+}
+
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
 	return 0;
