@@ -27,24 +27,30 @@
 #include <unistd.h>
 #include <string.h>
 #include <sys/socket.h>
+#include <sys/ioctl.h>
 #include <arpa/inet.h>
 #include <netinet/ether.h>
+#include <netinet/icmp6.h>
 #include <net/if_arp.h>
 #include <linux/if.h>
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
+#include <linux/wireless.h>
 
 #include <glib.h>
 
 #include "connman.h"
 
+#ifndef ARPHDR_PHONET_PIPE
+#define ARPHDR_PHONET_PIPE (821)
+#endif
+
 #define print(arg...) do { } while (0)
 //#define print(arg...) connman_info(arg)
 
 struct watch_data {
 	unsigned int id;
 	int index;
-	connman_rtnl_operstate_cb_t operstate;
 	connman_rtnl_link_cb_t newlink;
 	void *user_data;
 };
@@ -60,7 +66,8 @@ struct interface_data {
 	int index;
 	char *name;
 	char *ident;
-	enum connman_service_type type;
+	enum connman_service_type service_type;
+	enum connman_device_type device_type;
 };
 
 static GHashTable *interface_list = NULL;
@@ -69,7 +76,7 @@ static void free_interface(gpointer data)
 {
 	struct interface_data *interface = data;
 
-	__connman_technology_remove_interface(interface->type,
+	__connman_technology_remove_interface(interface->service_type,
 			interface->index, interface->name, interface->ident);
 
 	g_free(interface->ident);
@@ -90,18 +97,48 @@ static connman_bool_t ether_blacklisted(const char *name)
 	if (g_str_has_prefix(name, "vboxnet") == TRUE)
 		return TRUE;
 
+	/* virtual interface from Virtual Machine Manager */
+	if (g_str_has_prefix(name, "virbr") == TRUE)
+		return TRUE;
+
 	return FALSE;
 }
 
+static connman_bool_t wext_interface(char *ifname)
+{
+	struct iwreq wrq;
+	int fd, err;
+
+	fd = socket(PF_INET, SOCK_DGRAM, 0);
+	if (fd < 0)
+		return FALSE;
+
+	memset(&wrq, 0, sizeof(wrq));
+	strncpy(wrq.ifr_name, ifname, IFNAMSIZ);
+
+	err = ioctl(fd, SIOCGIWNAME, &wrq);
+
+	close(fd);
+
+	if (err < 0)
+		return FALSE;
+
+	return TRUE;
+}
+
 static void read_uevent(struct interface_data *interface)
 {
 	char *filename, line[128];
+	connman_bool_t found_devtype;
 	FILE *f;
 
-	if (ether_blacklisted(interface->name) == TRUE)
-		interface->type = CONNMAN_SERVICE_TYPE_UNKNOWN;
-	else
-		interface->type = CONNMAN_SERVICE_TYPE_ETHERNET;
+	if (ether_blacklisted(interface->name) == TRUE) {
+		interface->service_type = CONNMAN_SERVICE_TYPE_UNKNOWN;
+		interface->device_type = CONNMAN_DEVICE_TYPE_UNKNOWN;
+	} else {
+		interface->service_type = CONNMAN_SERVICE_TYPE_ETHERNET;
+		interface->device_type = CONNMAN_DEVICE_TYPE_ETHERNET;
+	}
 
 	filename = g_strdup_printf("/sys/class/net/%s/uevent",
 						interface->name);
@@ -113,6 +150,7 @@ static void read_uevent(struct interface_data *interface)
 	if (f == NULL)
 		return;
 
+	found_devtype = FALSE;
 	while (fgets(line, sizeof(line), f)) {
 		char *pos;
 
@@ -124,58 +162,55 @@ static void read_uevent(struct interface_data *interface)
 		if (strncmp(line, "DEVTYPE=", 8) != 0)
 			continue;
 
-		if (strcmp(line + 8, "wlan") == 0)
-			interface->type = CONNMAN_SERVICE_TYPE_WIFI;
-		else if (strcmp(line + 8, "wwan") == 0)
-			interface->type = CONNMAN_SERVICE_TYPE_CELLULAR;
-		else if (strcmp(line + 8, "bluetooth") == 0)
-			interface->type = CONNMAN_SERVICE_TYPE_BLUETOOTH;
-		else if (strcmp(line + 8, "wimax") == 0)
-			interface->type = CONNMAN_SERVICE_TYPE_WIMAX;
-		else
-			interface->type = CONNMAN_SERVICE_TYPE_UNKNOWN;
+		found_devtype = TRUE;
+
+		if (strcmp(line + 8, "wlan") == 0) {
+			interface->service_type = CONNMAN_SERVICE_TYPE_WIFI;
+			interface->device_type = CONNMAN_DEVICE_TYPE_WIFI;
+		} else if (strcmp(line + 8, "wwan") == 0) {
+			interface->service_type = CONNMAN_SERVICE_TYPE_CELLULAR;
+			interface->device_type = CONNMAN_DEVICE_TYPE_CELLULAR;
+		} else if (strcmp(line + 8, "bluetooth") == 0) {
+			interface->service_type = CONNMAN_SERVICE_TYPE_BLUETOOTH;
+			interface->device_type = CONNMAN_DEVICE_TYPE_BLUETOOTH;
+		} else if (strcmp(line + 8, "wimax") == 0) {
+			interface->service_type = CONNMAN_SERVICE_TYPE_WIMAX;
+			interface->device_type = CONNMAN_DEVICE_TYPE_WIMAX;
+		} else if (strcmp(line + 8, "gadget") == 0) {
+			interface->service_type = CONNMAN_SERVICE_TYPE_GADGET;
+			interface->device_type = CONNMAN_DEVICE_TYPE_GADGET;
+
+		} else {
+			interface->service_type = CONNMAN_SERVICE_TYPE_UNKNOWN;
+			interface->device_type = CONNMAN_DEVICE_TYPE_UNKNOWN;
+		}
 	}
 
 	fclose(f);
+
+	if (found_devtype)
+		return;
+
+	/* We haven't got a DEVTYPE, let's check if it's a wireless device */
+	if (wext_interface(interface->name)) {
+		interface->service_type = CONNMAN_SERVICE_TYPE_WIFI;
+		interface->device_type = CONNMAN_DEVICE_TYPE_WIFI;
+
+		connman_error("%s runs an unsupported 802.11 driver",
+				interface->name);
+	}
 }
 
-/**
- * connman_rtnl_add_operstate_watch:
- * @index: network device index
- * @callback: callback function
- * @user_data: callback data;
- *
- * Add a new RTNL watch for operation state events
- *
- * Returns: %0 on failure and a unique id on success
- */
-unsigned int connman_rtnl_add_operstate_watch(int index,
-			connman_rtnl_operstate_cb_t callback, void *user_data)
+enum connman_device_type __connman_rtnl_get_device_type(int index)
 {
-	struct watch_data *watch;
+	struct interface_data *interface;
 
-	watch = g_try_new0(struct watch_data, 1);
-	if (watch == NULL)
-		return 0;
+	interface = g_hash_table_lookup(interface_list,
+					GINT_TO_POINTER(index));
+	if (interface == NULL)
+		return CONNMAN_DEVICE_TYPE_UNKNOWN;
 
-	watch->id = ++watch_id;
-	watch->index = index;
-
-	watch->operstate = callback;
-	watch->user_data = user_data;
-
-	watch_list = g_slist_prepend(watch_list, watch);
-
-	DBG("id %d", watch->id);
-
-	if (callback) {
-		unsigned char operstate = 0;
-
-		if (operstate > 0)
-			callback(operstate, user_data);
-	}
-
-	return watch->id;
+	return interface->device_type;
 }
 
 /**
@@ -371,6 +406,7 @@ static void process_newlink(unsigned short type, int index, unsigned flags,
 	struct ether_addr compare = {{ 0, 0, 0, 0, 0, 0 }};
 	struct rtnl_link_stats stats;
 	unsigned char operstate = 0xff;
+	struct interface_data *interface;
 	const char *ifname = NULL;
 	unsigned int mtu = 0;
 	char ident[13], str[18];
@@ -398,6 +434,7 @@ static void process_newlink(unsigned short type, int index, unsigned flags,
 	switch (type) {
 	case ARPHRD_ETHER:
 	case ARPHRD_LOOPBACK:
+	case ARPHDR_PHONET_PIPE:
 	case ARPHRD_NONE:
 		__connman_ipconfig_newlink(index, type, flags,
 							str, mtu, &stats);
@@ -413,10 +450,8 @@ static void process_newlink(unsigned short type, int index, unsigned flags,
 						ifname, index, operstate,
 						operstate2str(operstate));
 
-	if (g_hash_table_lookup(interface_list,
-					GINT_TO_POINTER(index)) == NULL) {
-		struct interface_data *interface;
-
+	interface = g_hash_table_lookup(interface_list, GINT_TO_POINTER(index));
+	if (interface == NULL) {
 		interface = g_new0(struct interface_data, 1);
 		interface->index = index;
 		interface->name = g_strdup(ifname);
@@ -428,7 +463,7 @@ static void process_newlink(unsigned short type, int index, unsigned flags,
 		if (type == ARPHRD_ETHER)
 			read_uevent(interface);
 
-		__connman_technology_add_interface(interface->type,
+		__connman_technology_add_interface(interface->service_type,
 			interface->index, interface->name, interface->ident);
 	}
 
@@ -445,9 +480,6 @@ static void process_newlink(unsigned short type, int index, unsigned flags,
 		if (watch->index != index)
 			continue;
 
-		if (operstate != 0xff && watch->operstate)
-			watch->operstate(operstate, watch->user_data);
-
 		if (watch->newlink)
 			watch->newlink(flags, change, watch->user_data);
 	}
@@ -469,16 +501,6 @@ static void process_dellink(unsigned short type, int index, unsigned flags,
 						ifname, index, operstate,
 						operstate2str(operstate));
 
-	for (list = watch_list; list; list = list->next) {
-		struct watch_data *watch = list->data;
-
-		if (watch->index != index)
-			continue;
-
-		if (operstate != 0xff && watch->operstate)
-			watch->operstate(operstate, watch->user_data);
-	}
-
 	for (list = rtnl_list; list; list = list->next) {
 		struct connman_rtnl *rtnl = list->data;
 
@@ -1002,6 +1024,145 @@ static void rtnl_delroute(struct nlmsghdr *hdr)
 						msg, RTM_PAYLOAD(hdr));
 }
 
+static void *rtnl_nd_opt_rdnss(struct nd_opt_hdr *opt, guint32 *lifetime,
+			       int *nr_servers)
+{
+	guint32 *optint = (void *)opt;
+
+	if (opt->nd_opt_len < 3)
+		return NULL;
+
+	if (*lifetime > ntohl(optint[1]))
+		*lifetime = ntohl(optint[1]);
+
+	/* nd_opt_len is in units of 8 bytes. The header is 1 unit (8 bytes)
+	   and each address is another 2 units (16 bytes).
+	   So the number of addresses (given rounding) is nd_opt_len/2 */
+	*nr_servers = opt->nd_opt_len / 2;
+
+	/* And they start 8 bytes into the packet, or two guint32s in. */
+	return optint + 2;
+}
+
+static const char **rtnl_nd_opt_dnssl(struct nd_opt_hdr *opt, guint32 *lifetime)
+{
+	const char **domains = NULL;
+	guint32 *optint = (void *)opt;
+	unsigned char *optc = (void *)&optint[2];
+	int data_len = (opt->nd_opt_len * 8) - 8;
+	int nr_domains = 0;
+	int i, tmp;
+
+	if (*lifetime > ntohl(optint[1]))
+		*lifetime = ntohl(optint[1]);
+
+	/* Turn it into normal strings by converting the length bytes into '.',
+	   and count how many search domains there are while we're at it. */
+	i = 0;
+	while (i < data_len) {
+		if (optc[i] > 0x3f) {
+			DBG("DNSSL contains compressed elements in violation of RFC6106");
+			return NULL;
+		}
+
+		if (optc[i] == 0) {
+			nr_domains++;
+			i++;
+			/* Check for double zero */
+			if (i < data_len && optc[i] == 0)
+				break;
+			continue;
+		}
+
+		tmp = i;
+		i += optc[i] + 1;
+
+		if (i >= data_len) {
+			DBG("DNSSL data overflows option length");
+			return NULL;
+		}
+
+		optc[tmp] = '.';
+	}
+
+	domains = g_try_new0(const char *, nr_domains + 1);
+	if (!domains)
+		return NULL;
+
+	/* Now point to the normal strings, missing out the leading '.' that
+	   each of them will have now. */
+	for (i = 0; i < nr_domains; i++) {
+		domains[i] = (char *)optc + 1;
+		optc += strlen((char *)optc) + 1;
+	}
+
+	return domains;
+}
+
+static void rtnl_newnduseropt(struct nlmsghdr *hdr)
+{
+	struct nduseroptmsg *msg = (struct nduseroptmsg *) NLMSG_DATA(hdr);
+	struct nd_opt_hdr *opt = (void *)&msg[1];
+	guint32 lifetime = -1;
+	const char **domains = NULL;
+	struct in6_addr *servers;
+	int nr_servers = 0;
+	int msglen = msg->nduseropt_opts_len;
+	char *interface;
+
+	DBG("family %02x index %x len %04x type %02x code %02x",
+	    msg->nduseropt_family, msg->nduseropt_ifindex,
+	    msg->nduseropt_opts_len, msg->nduseropt_icmp_type,
+	    msg->nduseropt_icmp_code);
+
+	if (msg->nduseropt_family != AF_INET6 ||
+	    msg->nduseropt_icmp_type != ND_ROUTER_ADVERT ||
+	    msg->nduseropt_icmp_code != 0)
+		return;
+
+	interface = connman_inet_ifname(msg->nduseropt_ifindex);
+	if (!interface)
+		return;
+
+	for (opt = (void *)&msg[1];
+	     msglen >= 2 && msglen >= opt->nd_opt_len && opt->nd_opt_len;
+	     msglen -= opt->nd_opt_len,
+		     opt = ((void *)opt) + opt->nd_opt_len*8) {
+
+		DBG("nd opt type %d len %d\n",
+		    opt->nd_opt_type, opt->nd_opt_len);
+
+		if (opt->nd_opt_type == 25)
+			servers = rtnl_nd_opt_rdnss(opt, &lifetime,
+						    &nr_servers);
+		else if (opt->nd_opt_type == 31)
+			domains = rtnl_nd_opt_dnssl(opt, &lifetime);
+	}
+
+	if (nr_servers) {
+		int i, j;
+		char buf[40];
+
+		for (i = 0; i < nr_servers; i++) {
+			if (!inet_ntop(AF_INET6, servers + i, buf, sizeof(buf)))
+				continue;
+
+			if (domains == NULL || domains[0] == NULL) {
+				connman_resolver_append_lifetime(interface,
+							NULL, buf, lifetime);
+				continue;
+			}
+
+			for (j = 0; domains[j]; j++)
+				connman_resolver_append_lifetime(interface,
+								domains[j],
+								buf, lifetime);
+		}
+	}
+	g_free(domains);
+	g_free(interface);
+}
+
 static const char *type2string(uint16_t type)
 {
 	switch (type) {
@@ -1029,6 +1190,8 @@ static const char *type2string(uint16_t type)
 		return "NEWROUTE";
 	case RTM_DELROUTE:
 		return "DELROUTE";
+	case RTM_NEWNDUSEROPT:
+		return "NEWNDUSEROPT";
 	default:
 		return "UNKNOWN";
 	}
@@ -1153,6 +1316,9 @@ static void rtnl_message(void *buf, size_t len)
 		case RTM_DELROUTE:
 			rtnl_delroute(hdr);
 			break;
+		case RTM_NEWNDUSEROPT:
+			rtnl_newnduseropt(hdr);
+			break;
 		}
 
 		len -= hdr->nlmsg_len;
@@ -1165,17 +1331,22 @@ static gboolean netlink_event(GIOChannel *chan,
 {
 	unsigned char buf[4096];
 	gsize len;
-	GIOError err;
+	GIOStatus status;
 
 	if (cond & (G_IO_NVAL | G_IO_HUP | G_IO_ERR))
 		return FALSE;
 
 	memset(buf, 0, sizeof(buf));
 
-	err = g_io_channel_read(chan, (gchar *) buf, sizeof(buf), &len);
-	if (err) {
-		if (err == G_IO_ERROR_AGAIN)
-			return TRUE;
+	status = g_io_channel_read_chars(chan, (gchar *) buf,
+						sizeof(buf), &len, NULL);
+
+	switch (status) {
+	case G_IO_STATUS_NORMAL:
+		break;
+	case G_IO_STATUS_AGAIN:
+		return TRUE;
+	default:
 		return FALSE;
 	}
 
@@ -1333,7 +1504,8 @@ int __connman_rtnl_init(void)
 	memset(&addr, 0, sizeof(addr));
 	addr.nl_family = AF_NETLINK;
 	addr.nl_groups = RTMGRP_LINK | RTMGRP_IPV4_IFADDR | RTMGRP_IPV4_ROUTE |
-				RTMGRP_IPV6_IFADDR | RTMGRP_IPV6_ROUTE;
+				RTMGRP_IPV6_IFADDR | RTMGRP_IPV6_ROUTE |
+				(1<<(RTNLGRP_ND_USEROPT-1));
 
 	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 		close(sk);
@@ -1343,6 +1515,9 @@ int __connman_rtnl_init(void)
 	channel = g_io_channel_unix_new(sk);
 	g_io_channel_set_close_on_unref(channel, TRUE);
 
+	g_io_channel_set_encoding(channel, NULL, NULL);
+	g_io_channel_set_buffered(channel, FALSE);
+
 	g_io_add_watch(channel, G_IO_IN | G_IO_NVAL | G_IO_HUP | G_IO_ERR,
 							netlink_event, NULL);
 
